/*
 * @(#)ZipInputStream.java
 *
 * (c) Copyright Sun Microsystems Inc, 1992-1999. All rights reserved.
 * (c) Copyright IBM Corporation, 1996-1999. All rights reserved.
 * 
 * This product is based in part on the work of the Independent JPEG group.
 * 
 * The Java(tm) technology is owned and exclusively licensed by 
 * Sun Microsystems Inc.  Java is a trademark of Sun Microsystems Inc. in
 * the U.S. and other countries.
 * 
 * US Govt Users Restricted Rights - Use duplication or disclosure 
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 */
 
/*
 * @(#)ZipInputStream.java	1.11 97/01/24
 * 
 * Copyright (c) 1995, 1996 Sun Microsystems, Inc. All Rights Reserved.
 * 
 */

package java.util.zip;

import java.io.InputStream;
import java.io.IOException;
import java.io.EOFException;
import java.io.PushbackInputStream;

/**
 * This class implements an input stream filter for reading files in the
 * ZIP file format. Includes support for both compressed and uncompressed
 * entries.
 *
 * @author	David Connelly
 * @version	1.11, 01/24/97
 */
public
class ZipInputStream extends InflaterInputStream implements ZipConstants {
 
    private static final String copyr = 
        "Licensed Materials - Property of IBM\n" +
        "IBM Developer Kit, Java(TM) Tech Edition\n" +
        "(c) Copyright IBM Corp. 1996, 1999. All rights reserved.\n" +
        "\n" +
        "US Government Users Restricted Rights - Use,\n" +
        "duplication or disclosure restricted by GSA\n" +
        "ADP Schedule Contract with IBM Corp.\n" ;
 
    private ZipEntry entry;
    private CRC32 crc = new CRC32();
    private long remaining;
    private byte[] tmpbuf = new byte[512];

    private static final int STORED = ZipEntry.STORED;
    private static final int DEFLATED = ZipEntry.DEFLATED;

    /**
     * Creates a new ZIP input stream.
     * @param in the actual input stream
     */
    public ZipInputStream(InputStream in) {
	super(new PushbackInputStream(in, 512), new Inflater(true), 512);
    }

    /**
     * Reads the next ZIP file entry and positions stream at the beginning
     * of the entry data.
     * @exception ZipException if a ZIP file error has occurred
     * @exception IOException if an I/O error has occurred
     */
    public ZipEntry getNextEntry() throws IOException {
	if (entry != null) {
	    closeEntry();
	}
	crc.reset();
	inf.reset();
	if ((entry = readLOC()) == null) {
	    return null;
	}
	if (entry.method == STORED) {
	    remaining = entry.size;
	}
	return entry;
    }

    /**
     * Closes the current ZIP entry and positions the stream for reading the
     * next entry.
     * @exception ZipException if a ZIP file error has occurred
     * @exception IOException if an I/O error has occurred
     */
    public void closeEntry() throws IOException {
	while (read(tmpbuf, 0, tmpbuf.length) != -1) ;
    }
    
    /**
     * Reads from the current ZIP entry into an array of bytes. Blocks until
     * some input is available.
     * @param b the buffer into which the data is read
     * @param off the start offset of the data
     * @param len the maximum number of bytes read
     * @return the actual number of bytes read, or -1 if the end of the
     *         entry is reached
     * @exception ZipException if a ZIP file error has occurred
     * @exception IOException if an I/O error has occurred
     */
    public int read(byte[] b, int off, int len) throws IOException {
	if (entry == null) {
	    return -1;
	}
	switch (entry.method) {
	case DEFLATED:
	    len = super.read(b, off, len);
	    if (len == -1) {
		readEnd(entry);
		entry = null;
	    } else {
		crc.update(b, off, len);
	    }
	    return len;
	case STORED:
	    if (remaining <= 0) {
                if ((entry.flag & 8) == 8) {                                        /*ibm.9744*/
                    /* EXT descriptor present */                                    /*ibm.9744*/
                    readFully(tmpbuf, 0, EXTHDR);                                   /*ibm.9744*/
                }                                                                   /*ibm.9744*/
		entry = null;
		return -1;
	    }
	    if (len > remaining) {
		len = (int)remaining;
	    }
	    len = in.read(b, off, len);
	    if (len == -1) {
		throw new ZipException("unexpected EOF");
	    }
	    crc.update(b, off, len);
	    remaining -= len;
	    return len;
	default:
	    throw new InternalError("invalid compression method");
	}
    }

    /**
     * Skips specified number of bytes in the current ZIP entry.
     * @param n the number of bytes to skip
     * @return the actual number of bytes skipped
     * @exception ZipException if a ZIP file error has occurred
     * @exception IOException if an I/O error has occurred
     */
    public long skip(long n) throws IOException {
	if (n <= 0) {
	    return 0;
	}
	n = Math.min(n, Integer.MAX_VALUE);
	int total = 0;
	while (total < n) {
	    int len = read(tmpbuf, 0, (int)n - total);
	    if (len == -1) {
		break;
	    }
	    total += len;
	}
	return total;
    }

    /**
     * Closes the ZIP input stream.
     * @exception IOException if an I/O error has occurred
     */
    public void close() throws IOException {
	in.close();
    }

    /*
     * Reads local file (LOC) header for next entry.
     */
    private ZipEntry readLOC() throws IOException {
	try {
	    readFully(tmpbuf, 0, LOCHDR);
	} catch (EOFException e) {
	    return null;
	}
	if (get32(tmpbuf, 0) != LOCSIG) {
	    return null;
	}


        // IBM.11259 - This is part of the fix for 10089 that did not get put in under that defect
        // The defect fix was in two parts - ZipOutputStream.java and ZipInputStream.java.  
        // The changes are to make sure that the names are written out in UTF-8 instead
        // of ASCII.  This is to ensure that DBCS filenames are maintained.

        // original line of code
	//ZipEntry e = new ZipEntry();                                              //IBM.11259

	// get the entry name and create the ZipEntry first                      //ibm.11259
	int len = get16(tmpbuf, LOCNAM);                                         //ibm.11259
	if (len == 0) {                                                          //ibm.11259
	    throw new ZipException("missing entry name");                        //ibm.11259 
	}                                                                        //ibm.11259
        byte[] b = new byte[len];                                                //ibm.11259
	readFully(b, 0, len);                                                    //ibm.11259
	ZipEntry e = createZipEntry(getUTF8String(b, 0, len));                   //ibm.11259
	// now get the remaining fields for the entry                            //ibm.11259
                                                                                 //ibm.11259
        // ibm.11259 end


	e.version = get16(tmpbuf, LOCVER);
	e.flag = get16(tmpbuf, LOCFLG);
	if ((e.flag & 1) == 1) {
	    throw new ZipException("encrypted ZIP entry not supported");
	}
	e.method = get16(tmpbuf, LOCHOW);
	e.time = get32(tmpbuf, LOCTIM);
//	if ((e.flag & 8) == 8) {                                  /*ibm.9744*/
//	    /* EXT descriptor present */                          /*ibm.9744*/
//	    if (e.method != DEFLATED) {                           /*ibm.9744*/
//		throw new ZipException(                           /*ibm.9744*/
//		 "only DEFLATED entries can have EXT descriptor");/*ibm.9744*/
//	    }                                                     /*ibm.9744*/
//	} else {                                                  /*ibm.9744*/
//	    e.crc = get32(tmpbuf, LOCCRC);                        /*ibm.9744*/
//	    e.csize = get32(tmpbuf, LOCSIZ);                      /*ibm.9744*/
//	    e.size = get32(tmpbuf, LOCLEN);                       /*ibm.9744*/
//	}                                                         /*ibm.9744*/
	long size = get32(tmpbuf, LOCLEN); /* ibm.11613 */
	if( size>0 || e.isDirectory() ||  ((e.flag & 8) != 8)) {                         /* ibm.11613, ibm.12478 ibm.12527*/
	    e.crc = get32(tmpbuf, LOCCRC);                        /*ibm.9744*/
	    e.csize = get32(tmpbuf, LOCSIZ);                      /*ibm.9744*/
	    e.size = size;                                /*ibm.9744,11613*/
	} else if ((e.flag & 8) == 8) {                   /*ibm.9744,11613*/
	    /* EXT descriptor present */                          /*ibm.9744*/
	    if (e.method != DEFLATED){                    /*ibm.9744, 11613*/
		throw new ZipException(                           /*ibm.9744*/
		 "only DEFLATED entries can have EXT descriptor");/*ibm.9744*/
            }                                                     /*ibm.9744*/
        }                                                         /*ibm.9744*/

        // code moved high up 
        /*int len = get16(tmpbuf, LOCNAM);                                          //IBM.11259    
	if (len == 0) {                                                             //IBM.11259
	    throw new ZipException("missing entry name");                           //IBM.11259
	}                                                                           //IBM.11259
	byte[] b = new byte[len];                                                   //IBM.11259
	readFully(b, 0, len);                                                       //IBM.11259
	e.name = new String(b, 0, 0, len); */                                       //IBM.11259
	len = get16(tmpbuf, LOCEXT);
	if (len > 0) {
	    b = new byte[len];
	    readFully(b, 0, len);
	    e.extra = b;
	}
	return e;
    }


    // new methods for ibm.11259

    /*
     * Fetches a UTF8-encoded String from the specified byte array.
     */
    private static String getUTF8String(byte[] b, int off, int len) {
	// First, count the number of characters in the sequence
	int count = 0;
	int max = off + len;
	int i = off;
	while (i < max) {
	    int c = b[i++] & 0xff;
	    switch (c >> 4) {
	    case 0: case 1: case 2: case 3: case 4: case 5: case 6: case 7:
		// 0xxxxxxx
		count++;
		break;
	    case 12: case 13:
		// 110xxxxx 10xxxxxx
		if ((int)(b[i++] & 0xc0) != 0x80) {
		    throw new IllegalArgumentException();
		}
		count++;
		break;
	    case 14:
		// 1110xxxx 10xxxxxx 10xxxxxx
		if (((int)(b[i++] & 0xc0) != 0x80) ||
		    ((int)(b[i++] & 0xc0) != 0x80)) {
		    throw new IllegalArgumentException();
		}
		count++;
		break;
	    default:
		// 10xxxxxx, 1111xxxx
		throw new IllegalArgumentException();
	    }
	}
	if (i != max) {
	    throw new IllegalArgumentException();
	}
	// Now decode the characters...
	char[] cs = new char[count];
	i = 0;
	while (off < max) {
	    int c = b[off++] & 0xff;
	    switch (c >> 4) {
	    case 0: case 1: case 2: case 3: case 4: case 5: case 6: case 7:
		// 0xxxxxxx
		cs[i++] = (char)c;
		break;
	    case 12: case 13:
		// 110xxxxx 10xxxxxx
		cs[i++] = (char)(((c & 0x1f) << 6) | (b[off++] & 0x3f));
		break;
	    case 14:
		// 1110xxxx 10xxxxxx 10xxxxxx
		int t = (b[off++] & 0x3f) << 6;
		cs[i++] = (char)(((c & 0x0f) << 12) | t | (b[off++] & 0x3f));
		break;
	    default:
		// 10xxxxxx, 1111xxxx
		throw new IllegalArgumentException();
	    }
	}
	return new String(cs, 0, count);
    }

    /**
     * Creates a new <code>ZipEntry</code> object for the specified
     * entry name.
     *
     * @param name the ZIP file entry name
     */
    protected ZipEntry createZipEntry(String name) {
	return new ZipEntry(name);
    }


    // end of new methods for ibm.11259





    /*
     * Reads end of deflated entry as well as EXT descriptor if present.
     */
    private void readEnd(ZipEntry e) throws IOException {
	int n = inf.getRemaining();
	if (n > 0) {
	    ((PushbackInputStream)in).unread(buf, len - n, n);
	}
	if ((e.flag & 8) == 8) {
	    /* EXT descriptor present */
	    readFully(tmpbuf, 0, EXTHDR);
	    long sig = get32(tmpbuf, 0);
	    if (sig != EXTSIG) {
		throw new ZipException("invalid EXT descriptor signature");
	    }
	    e.crc = get32(tmpbuf, EXTCRC);
	    e.csize = get32(tmpbuf, EXTSIZ);
	    e.size = get32(tmpbuf, EXTLEN);
	}
	if (e.size != inf.getTotalOut()) {
	    throw new ZipException(
		"invalid entry size (expected " + e.size + " but got " +
		inf.getTotalOut() + " bytes)");
	}
	if (e.csize != inf.getTotalIn()) {
	    throw new ZipException(
		"invalid entry compressed size (expected " + e.csize +
		" but got " + inf.getTotalIn() + " bytes)");
	}
	if (e.crc != crc.getValue()) {
	    throw new ZipException(
		"invalid entry CRC (expected 0x" + Long.toHexString(e.crc) +
		" but got 0x" + Long.toHexString(crc.getValue()) + ")");
	}
    }

    /*
     * Reads bytes, blocking until all bytes are read.
     */
    private void readFully(byte[] b, int off, int len) throws IOException {
	while (len > 0) {
	    int n = in.read(b, off, len);
	    if (n == -1) {
		throw new EOFException();
	    }
	    off += n;
	    len -= n;
	}
    }

    /*
     * Fetches unsigned 16-bit value from byte array at specified offset.
     * The bytes are assumed to be in Intel (little-endian) byte order.
     */
    private static final int get16(byte b[], int off) {
	return (b[off] & 0xff) | ((b[off+1] & 0xff) << 8);
    }

    /*
     * Fetches unsigned 32-bit value from byte array at specified offset.
     * The bytes are assumed to be in Intel (little-endian) byte order.
     */
    private static final long get32(byte b[], int off) {
	return get16(b, off) | ((long)get16(b, off+2) << 16);
    }
}
